#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <sys/io.h>
#include <sys/mman.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <linux/rtc.h>

#include "tsc.h"

unsigned long long __t_start, __t_stop;

static double cpuKHz = 0.0;


/* return the CPU clock frequency in KHz (as a double) */
static void calibrate_tsc_without_rtc(void)
{
	FILE *f;
	double cpuMHz = 0.0;
        fprintf(stderr, "Calibrating TSC with /proc/cpuinfo\n");
	f = popen("cat /proc/cpuinfo|grep 'cpu MHz'|cut -d':' -f2", "r");
	if (!f) {
		perror("/proc/cpuinfo");
		exit(1);
	}
	if (fscanf(f, "%lf", &cpuMHz) != 1) {
		fprintf(stderr, "Fatal error: Cannot read CPU clock frequency from /proc/cpuinfo\n");
		exit(1);
	}
	pclose(f);
	cpuKHz = cpuMHz * 1000.0;
	fprintf(stderr, "CPU clock frequency: %lf KHz\n", cpuKHz);
}

/* WARNING:
 *      code stolen from linux 2.6.13
 *      this code is _NOT_ SMP safe!
 *      it should be executed with interrupt disabled
 */

static inline void rep_nop(void)
{
        __asm__ __volatile__("rep;nop" : : : "memory");        
}

#define RTC_PORT(x) (0x70+(x))
#define RTC_FREQ_SELECT 10
#define RTC_UIP 0x80

static inline unsigned char rtc_cmos_read(unsigned char addr)
{
        unsigned char val;
        
        outb_p(addr, RTC_PORT(0));
        val = inb_p(RTC_PORT(1));
        return val;        
}

static void calibrate_tsc_with_rtc(int interval, int iter)
{
        int fd, i, secs;
        unsigned long long cycles[16];
        double tot= 0.0, tot2 = 0.0, avg, stdev;

        if (iter > 16)
                iter = 16;
        if (iter <= 0)
                iter = 1;
        
        fd = open("/dev/rtc", O_RDWR);
        if (fd == -1) {
fallback:
                perror("/dev/rtc");
                calibrate_tsc_without_rtc();
                return;
        }
        
        if (ioctl(fd, RTC_AIE_OFF) == -1)
                goto fallback;
        
        if (ioctl(fd, RTC_PIE_OFF) == -1)
                goto fallback;

        if (ioctl(fd, RTC_UIE_OFF) == -1)
                goto fallback;

        fprintf(stderr, "Calibrating the TSC with the RTC...\n");        

        

        fprintf(stderr, "WARNING! Your computer will be frozen for %d seconds\n", 
                        interval*iter);
        fflush(stderr);
        sleep(1);
        
        for (i=0; i<iter ; ++i) {
                secs = interval;
                start_timing(1); /* to clear interrupts */

                while (!(rtc_cmos_read(RTC_FREQ_SELECT) & RTC_UIP))
                                rep_nop();

                while(rtc_cmos_read(RTC_FREQ_SELECT) & RTC_UIP)
                                rep_nop();

                start_timing(0); /* to reset start time */

                do {
                        while (!(rtc_cmos_read(RTC_FREQ_SELECT) & RTC_UIP))
                                rep_nop();
                        while(rtc_cmos_read(RTC_FREQ_SELECT) & RTC_UIP)
                                rep_nop();
                } while (--secs != 0);

                stop_timing(1);

                cycles[i] = cycle_result();
        }
        
        for (i=0; i<iter; ++i) {
                avg = (double)cycles[i]/interval;
                tot += avg;
                tot2 += avg*avg;
        }
        
        avg = (double)tot / iter;
        stdev = sqrt((double)tot2 / iter - avg*avg);
        fprintf(stderr, "CPU clock frequency: %.0f Hz (+/- %.0f Hz)\n",
                        avg, 2*stdev);
        fprintf(stderr, "The TSC counter is increased once in %f nanoseconds\n",
                                1e9/avg);
        if (close(fd) == -1)
                perror("/dev/rtc");
        cpuKHz = avg / 1e3; 
}

void calibrate_tsc(int rtc)
{                
        if (iopl(3) == -1) {
		fprintf(stderr, "Fatal error: iopl(3) failed. (Run me as root!)\n");
		exit(1);
	}
	if (mlockall(MCL_CURRENT|MCL_FUTURE)) {
		fprintf(stderr, "Fatal error: cannot lock pages in RAM\n");
		exit(1);
	}
        if (rtc)
		calibrate_tsc_with_rtc(5, 3);
	else
		calibrate_tsc_without_rtc();
}


double cycles_to_time(unsigned long long c)
{
	return (double) c / cpuKHz; /* milliseconds */
}

double timing_result(void)
{
	return cycles_to_time(__t_stop - __t_start);
}

unsigned long long cycle_result(void)
{
	return __t_stop - __t_start;
}

